;;;
;;;
;;;
;;; Virtual Disk Drive Project - Apr 26, 2006
;;;
;;;    Altair DCDD support
;;;
;;;

#include "main.h"
#include "vd.h"
#include "utility.h"
#include "headmove.mac"
#include "waitP1.mac"
#include "disk.h"
#include "write.mac"
#include "sertrack.h"
		
;;; NOTES on Altair floppy
;;;
;;;     rotation = 360 RPM (60 Hz)
;;;	   holes = 32, plus one index
;;;   hole width = 1.8ms (1800us) -> 
;;;   hole delta = 5.2ms (5200us) --> 32 * 5.2 = 166.4ms = 60 Hz'ish
;;;   index hole = 2.6ms from START of sector hole (.8ms from end)
;;;    data rate = 250 kHz
;;;    Bit-cell  = 4us - includes clock AND data
;;;		 = 20 inst cycles
;;;    Bit-pulse = 1us
;;;		 = 5 inst cycles
;;;   write gate = enabled "almost" immediately after sector index hole - 
;;;		   only delay is the software noticing the start of
;;;		   a sector and comparing the sector number
;;; write gate length = 5.2ms (the entire length of a sector)
;;;   write DATA = 280us after start of sector (0's until then)
;;;    read data = 140us after start of sector - until this time the
;;;		   ability to read is blanked, afterwhich, the first
;;;		   one (1) bit is noticed as the real start of data
;;;
;;; THEREFORE:
;;;	- generate zeros after the start of a sector within 140us
;;; 	          (700 cycles - 35 bits - 4.4'ish bytes)
;;; 	  of the start of the sector - this is when reading starts
;;;	  Though it could be started much earlier.
;;;	- start the data at 280us (1400 cycles) - this is where the write
;;;	  would have normally begun (1400 cycles - 70 bits - 8.8'ish bytes)
;;;	- data length = (137 bytes * 32us/byte) = 4.384ms
;;;	- sector = 280us + 4384us = 4664us
;;;     - trailer needs to be = 5200 - 4664 = 536us = 16.75 bytes = 134 bits

;;; NOTE - there was a mod that was recommended in '78 that moved the
;;;		data closer to the middle of the sector space.  This
;;;		mod included moving the write data start to 389us and
;;;		the read data start to 214us
;;;		This obviously changes the calculation for end padding.
;;;		Either pre-mod or mod values can be used.

#define SECTORMOD78	
;;; #undef SECTORMOD78
	
;;; the following defines are used to calculate the end of sector padding
;;; and are in micro-seconds (us)

#define SECTOR_LENGTH_uS	d'5200'
	
#ifdef SECTORMOD78
#define		WRITESTART_uS	d'389'
#define		READSTART_uS	d'214'
#else
#define		WRITESTART_uS	d'280'
#define		READSTART_uS	d'140'
#endif

;;; WRITESTART - the number of cycles before a write will occur - data is
;;;		sent out after this number of cycles.  During the zeros
;;;		that precede this time, write-line monitoring is done.
;;;		Zeros are generated for this amount of time at the start of sector.
;;; READSTART - the number of cycles before the Altair will monitor read
;;;		data.  The zeros must start before this time.
;;; TRAILER - the number of cycles that should be added to the tail-end
;;;		of a sector to "fill-up" the sector write-time.
;;; HEADER_TRAILER_BYTES - the number of bytes left over in the header after
;;;		all sectors have been generated:
;;;		- 16 * 1 sector switch bytes = 16
;;;		- 32 * 7 data bytes = 224
;;;		- 1 trailer identification byte = 1

#define CYCLES_PER_uS		d'5'
#define BIT_CELL_uS		d'4'
#define WRITESTART		(WRITESTART_uS * CYCLES_PER_uS)
#define WRITESTART_BIT_CELLS	(WRITESTART_uS / BIT_CELL_uS)
#define READSTART		(READSTART_uS * CYCLES_PER_uS)
#define BYTE_uS			(d'4' * d'8')
#define DATA_uS			(d'137' * BYTE_uS)
#define SECTOR_TRAILER_uS	(SECTOR_LENGTH_uS - WRITESTART_uS - DATA_uS)
#define TRAILER			(SECTOR_TRAILER_uS * CYCLES_PER_uS)
#define TRAILER_BIT_CELLS	(SECTOR_TRAILER_uS / BIT_CELL_uS)

;;; HEADER - number of bytes/register being read and generated from the header block
;;; PRE_INDEX - number of bytes & register being read from data block BEFORE index pulse
;;; POST_INDEX - number of bytes & register being read from data block AFTER index pulse
	
#define HEADER_BYTES		d'7'
#define HEADER			WORK0
	
#define PRE_INDEX_BYTES		d'60'
#define PRE_INDEX		WORK1
	
#define POST_INDEX_BYTES	d'68'
#define POST_INDEX		ARG2

;;; IS_LAST_SECTOR - the register which is set to 1 if this is the last sector
;;;			NOTE - that each "sector" which is called from the high-
;;;			level track routine is actually 2 sectors.  So a sector
;;;			number of 16 is the "last" sector.
	
#define IS_LAST_SECTOR		ARG1

;;; ADATA and ACOUNT are defines for registers used in AltairByte
	
#define ADATA			WORK2
#define ACOUNT			WORK3

PROG1	CODE
	global AltairSector
	global Altair2Sector
	global AltairWriteSector
	
AltairSector:	
Altair2Sector:	
AltairWriteSector:	
	
#ifdef NOTDEF

	
;;;**********************************************************************
;;; AltairSector() & Altair2Sector()
;;;
;;; DESCR:	The general structure for an Altair sector is:
;;;		- 2 sectors of data are in each block, so when generating
;;;		  a "sector" in SVD-speak, 2 Altair sectors are generated.
;;;		- the header has data for 32 sectors, although those sectors
;;;		  take up only 16 SVD-blocks.
;;;		- the outside SVD sector switch notices that we're talking
;;;		  about an Altair sector and ends-up generating 2 sectors -
;;;		  only one sector switch byte for 2 sectors
;;;		- simply put, the structure for sectors is:
;;;			- each sector is 135 bytes (out of 137)
;;;			- 7 bytes in the header for each sector
;;;				followed by 128 bytes x 2 in the sector block
;;;			- the bytes are all in order, smoothly going from
;;;			  the header block to the data blocks
;;;			- header bytes used is 
;;;				(1 * 16) = 16 bytes sector type switch
;;;				(7 * 32) = 224 bytes of data
;;;				      15 = alignment padding as we go along
;;;					 = 16 + 224 + 15 + 1 trailer switch = 256
;;;			- at the end of each sector pair, an extra INC is done to
;;;			  keep data aligned - EXCEPT for the last pair
;;;		- the 2 byte balance (between 135 and 137) is added auto-
;;; 		  matically at the end of 135 bytes and is 0xff 0x00.
;;;		  The 0xff is the STOP for the altdos sector (track >= 6)
;;;		  and the 0x00 is the "unused" byte.
;;;		- This mechanism looks to work just fine for Altdos
;;;		- I am a bit concerned for CPM or any other OS that requires
;;;		  all 137 bytes.  CPM looks OK though.
;;;		- NOTE that the data in the sector blocks is appropriately
;;;		  positioned so that after reading header information the
;;;		  data is in the right spot
;;;		- NOTE that the index-hole is SUPPOSED to be 1.8ms wide,
;;;		  though the hardware only looks at its rising edge - so
;;;		  we take advantage of that fact and generate a short one.
;;;		- NOTE that headmoves are monitored - the problem is that
;;;		  by definition they can be between 200ns and 2us - HOWEVER,
;;;		  it looks like the HW does 500ns.  The dongle will need
;;;		  to stretch this pulse to something allowing monitoring
;;;		  far less actively.  Actually, the SVD needs to build in
;;;		  a 74221 (two 121's) to stretch both the write pulse and
;;;		  the head move pulse.
;;;**********************************************************************

Altair2Sector:	
	clrf		IS_LAST_SECTOR		; () defaults to NOT last sector

	incf		SECCOUNT		; () prepare for next sector

	call		AltairSector		; () first sector in data block
	SkipIfLow	WRITE_GATE		; () write condition noticed
	goto		AltairWriteSector	; () go to write data routine

	btfsc		SECCOUNT,d'4'		; () notices 'sector' 16
	bsf		IS_LAST_SECTOR,0	; ()   sets this for second logical block
	
	call		AltairSector		; () second sector in data block
	SkipIfLow	WRITE_GATE		; () write condition noticed
	goto		AltairWriteSector	; () go to write data routine

	btfss		SECCOUNT,d'4'		; () no alignment on LAST sector
	MemLowInc				; () alignment as we go along - except last sector
	
	PAGESEL		SerTrackSecDone
	goto		SerTrackSecDone		; (-13) Done with sector

		
AltairSector:
	Raise		INDEX_HOLE		; () the hard sector/index hole SHOULD be 1.8ms long,
						;    though it is allowed to be shorter

	movlw		WRITESTART_BIT_CELLS-1	; () generate this many zeros before data
	call		AltairZeros	; ()

	Lower		INDEX_HOLE	; (0) short sector pulse
	
	;; the WRITE_GATE is noticed, causing an early return during zero generation

	SkipIfLow	WRITE_GATE	; (1) write condition
	return				; (2) just return upon write
		
	;; at this point we're at (8) with no data bit required at (10)
	;; so we have 12 cycles before needing to generate the pulse of
	;; the clock in front of the actual data

	;; set-up the data to be used below
		
	movlw		HEADER_BYTES	; (3) number of bytes read/generated from header block
	movwf		HEADER		; (4)  register to hold this number
	
	movlw		PRE_INDEX_BYTES	; (5) bytes read/generated from data block BEFORE index pulse
	movwf		PRE_INDEX	; (6)  register to hold this number
	
	movlw		POST_INDEX_BYTES; (7) bytes read/generated from data block AFTER index pulse
	movwf		POST_INDEX	; (8)  register to hold this number

	;; now, assuming correct position of the data, send out 7 bytes from header

ASLoop1:
	Wait1				; (9)
	Raise		RDATA		; (0) generate first clock pulse
	MemReadByte			; (1)
	Wait1				; (2)
	call		AltairByte	; (3) called at (3) comes back at (3)
	MemLowInc			; (3)
	Lower		RDATA		; (5) 
	decfsz		HEADER		; (6)
	goto		ASLoop1		; (7)

	;; now flip to the data segment and do 128 bytes in total
	
	MemAddrInc	SECCOUNT	; (8) flip to data segment

	;; note that the data in the data block is interleaved ODD and EVEN
	;; that is, the first sector is considered EVEN and the second sector
	;; is considered ODD.  Each memory increment is done twice for each
	;; byte.  This is done so that after a 128-byte REAL sector
	;; we are positioned back to where we started, as is normally the case
	;; the second sector in the block.

	;; the first half of the sector (before a possible index pulse) is slid
	;; "back" by one cycle, allowing two increments.  If we didn't need
	;; two increments, the code would look like this:
	;;
	;; ASLoop2:	
	;; 	Raise		RDATA		; (0)
	;; 	MemReadByte			; (1)
	;; 	Wait1				; (2)
	;; 	call		AltairByte	; (3) called at (3) comes back at (3)
	;; 	MemLowInc			; (3)
	;; 	Lower		RDATA		; (5)
	;; 	Wait1				; (6)
	;; 	decfsz		PRE_INDEX	; (7)
	;; 	goto		ASLoop2		; (8)
	;; 

ASLoop2:	
	Raise		RDATA		; (0) this pulse is longer than normal (6 cycles)
	MemReadByte			; (1)
	MemLowInc			; (2)
	call		AltairByte	; (4) normally called at (3) coming back at (3) - slid out a bit internally
	MemLowInc			; (4)
	Lower		RDATA		; (6) keeps the normal size of the RDATA pulses
	decfsz		PRE_INDEX	; (7)
	goto		ASLoop2		; (8)

	;; generate the Index Pulse, and the rest of the data

	btfsc		IS_LAST_SECTOR,0; (9)
	Raise		INDEX_HOLE	; (0)
	
	;; OFF BY ONE AT THIS POINT - hummph...
	;; That is, we've taken one more cycle than we should have, which would cause
	;; us to slide out even more, so since we need two inc's anyway, we slide forward
	;; a bit
	
ASLoop3:	
	Raise		RDATA		; (0) this pulse is shorter than normal (4 cycles)
	MemReadByte			; (1)
	call		AltairByte	; (2) normally called at (3) coming back at (3) - slid in a bit internally
	MemLowInc			; (2)
	Lower		RDATA		; (4) keeps the normal size of the RDATA pulses
	MemLowInc			; (5)
	decfsz		POST_INDEX	; (7)
	goto		ASLoop3		; (8)

	Lower		INDEX_HOLE	; (9)

	;; generate two trailer bytes (0xff and 0x00)
	;; the first we do explicitely, and the second is simply part of the
	;; zeros we generate as part of a sector trailer

	Raise		RDATA		; (0)
	movlw		0xff		; (1)
	Wait1				; (2)
	call		AltairByte	; (3)

	;; flip back to the header
	
	MemAddrLoSet			; (3) flip back to header
	Lower		RDATA		; (5)

	;; generate zero trailer - note that unlike above where AltairZeros() is passed
	;;   the target number minus one, we use the whole number because we are generating
	;;   the last zero for the data as well.

	movlw		TRAILER_BIT_CELLS	; (6) generate this many zeros as trailer
	call		AltairZeros		; (7)
	
	return

;;;*******************************************************************
;;; NAME:	AltairZeros()
;;;
;;; DESCR:	Generates a squence of the given number of zeros,
;;;		which is simply a bit cell with a clock and no data.
;;;
;;; ARGS:	W - the number of zeros to generate - 1
;;;			In other words, W + 1 zeros are generated.
;;;			This makes return earlier.
;;;
;;; RETURNS:	the write and head-move are checked.
;;;
;;; NOTES:	- bit cells are 20 cycles wide (4us)
;;;		- called at (-3,7) returns at (0)
;;;		- NOTE - upon return we are at the end of the CLOCK pulse
;;;			and the next data period needs to be zero.  In
;;;			essence, you have a bunch of time on your hands.
;;;		- caller should check for the write gate upon return
;;;			if it is a condition of interest.
;;;*******************************************************************

AltairZeros:
	movwf		ACOUNT		; (-1)
	
AZLoop:	
	Raise		RDATA		; (0)
	Wait4				; (1) headcheck
	Lower		RDATA		; (5)

	decf		ACOUNT		; (6)
	skpnz				; (7)
	return				; (8) 

	Wait2				; (9) headcheck

	IfWriteReturn			; (1) upon write condition, return

	Wait3				; (5) headcheck

	goto		AZLoop		; (8)

;;;**********************************************************************
;;; NAME:	AltairByte()
;;;
;;; DESCR:	Does a 5-cycle wide data pulse for a byte given the
;;;		Altair 20-cycle (4us) full bit cell - that is, the
;;;		clock is 10 cycles and the data is 10 cycles.
;;;
;;;		RDATA needs to be brought up before the routine and
;;;		brought down after it.  Saves cycles.
;;;
;;; ARGS:	W has the byte to generate
;;;		ABYTE stores the data byte
;;;		ACOUNT stores a temporary count
;;;
;;; RETURNS:	
;;;
;;; NOTES:	- called at (3), returns at (3)
;;;		- does a bunch of head checks, essentially during data
;;;		  generation, a head check is done at LEAST once every
;;;		  three bits, or at least once in 12us
;;;**********************************************************************
AltairByte:
	;; upon entry the Raise has occurred at (0) and we are at (5)
	
	;; the caller has already raised RDATA, accounting for the first clock pulse.
	;; We have up to 15 more pulses to go.  The first clock pulse is generated outside
	;; of the loop and the last data pulse is generated outside the loop.

	Lower	RDATA		; (5) end of current clock pulse
	movwf	ADATA		; (6) get the W data into our working register
	movlw	0x07		; (7) prepare for 7 cycles of the loop
	movwf	ACOUNT		; (8)

AltairByteLoop:
	btfsc	ADATA,7		; (9) look right into the data
	Raise	RDATA		; (0) data bit here - UP
	Wait4			; (1)
	Lower	RDATA		; (5)               - DOWN
	Wait4			; (6)
	Raise	RDATA		; (0) clock bit here - UP
	Wait3			; (1)
	rlf	ADATA		; (4) shift data over
	Lower	RDATA		; (5)			- DOWN 
	decfsz	ACOUNT		; (6)
	goto	AltairByteLoop	; (7)

	;; at this point we need to generate a last data bit if required

	Wait1			; (8)
	btfsc	ADATA,7		; (9)
	Raise	RDATA		; (0)

	return			; (1)

AltairWriteSector:	

#endif
					
;;; =======================================================================
;;; =======================================================================

	END

